#pragma once
#include <cmath>
#include"scalar.h"
struct Vec3;

Scalar dot(const Vec3 &a, const Vec3 &b);

Vec3 cross(const Vec3 &a, const Vec3 &b);



Scalar degrees_between(const Vec3 &a, const Vec3 &b);

struct Vec3 {
    Scalar x;   
    Scalar y;
    Scalar z;

    Vec3(Scalar _x, Scalar _y, Scalar _z) : x(_x), y(_y), z(_z) {}  //per inizializzare i campi nel costruttore
    
    Vec3() : Vec3(0,0,0) {} //invochiamo l'altro costruttore = Vec3(): x(0), y(0), z(0) {}

    void scale(Scalar k) {
        this->x *= k;         
        y *= k;
        z *= k;
    }

    Vec3 scaled(Scalar k) const { //copia del Vec3 su cui viene chiamata la funzione
        return Vec3(x * k, y * k, z * k);
    }

    // Il passaggio di parametri è fatto sempre per copia, e gli oggetti pure vengono copiati.
    // Per passare per riferimento bisogna mettere & prima del nome del parametro (Vec3 &other): qui conviene farlo per risparmiare memoria!
    // Per assicurare al chiamante che anche se me lo faccio passare per riferimento non lo cambierò metto const davanti al nome -> il compilatore mi controlla!
    void add(const Vec3 &other) {
        x += other.x;
        y += other.y;
        z += other.z;
    }

    Vec3 added(const Vec3 &other) const {  
        return Vec3(x + other.x, y + other.y, z + other.z);
    }

    Scalar length() const { //dot(v,v) = magnitude(v)^2
        return std::sqrt(dot(*this,*this));
    }

    void normalize() {  
        scale(1 / length());     //*this = l'oggetto puntato da this, che è me stesso
    }

    Vec3 normalized() {
        return scaled(1 / length());
    }

    Vec3 operator *(Scalar k) const {     // Scrivendo operator * stiamo ridefinendo il * per i Vec3: però non posso inventare nuovi operatori, non posso cambiare il numero di parametri e neanche l'ordine di applicazione! (* sarà sempre prima di +)
        return Vec3(x * k, y * k, z * k);
    }

    void operator *=(Scalar k) {
        x *= k;          
        y *= k;
        z *= k;
    }

    Vec3 operator +(const Vec3 &other) const {  
        return Vec3(x + other.x, y + other.y, z + other.z);
    }

    void operator +=(const Vec3 &other) {
        x += other.x;
        y += other.y;
        z += other.z;
    }

    Vec3 operator -() const {
        return Vec3(-x,-y,-z);
    }

    Vec3 operator -(const Vec3 &other) const {
        return Vec3(x - other.x, y - other.y, z - other.z);
    }

    void operator -=(const Vec3 &other) {
        x -= other.x;
        y -= other.y;
        z -= other.z;
    }


    bool operator ==(const Vec3 &other) const {
        return x == other.x && y == other.y && z == other.z;    //Ordine operatori: unari -> binari, comparazione (==, <...) -> booleani
    }

    void printf() const;

    // friend dot -> Indico che la funz. globale dot può manipolare eventuali attributi privati di Vec3 
};

inline Vec3 operator *(Scalar k, const Vec3 &v) {
    return v*k;
}

inline Vec3 mix(const Vec3 &a, const Vec3 &b, Scalar t) {
    return (1-t)*a + t*b; //1-t e non 1.0 perchè il casting lo fa da solo
}

inline Vec3 nlerp(const Vec3 &a, const Vec3 &b, Scalar t) {
    return mix(a, b, t).normalized();
}
